/*
 * rule.c - Bytecode compiler and interpreter for rules
 *
 * This file is part of tcpspy, a TCP/IP connection monitor. 
 *
 * Copyright (c) 2000 Tim J. Robbins. 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: rule.c,v 1.1 2000/11/27 12:32:51 fyre Exp $
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

/*
 * These header files contain the data types used in "grammar.h".
 */
#include <sys/types.h>
#include <unistd.h>
#include <netinet/in.h>

#include "rule_grammar.h"
#include "rule.h"

typedef short unsigned int bytecode_t;
typedef unsigned char bool_t;

/*
 * Exported functions.
 */
void rule_gen_user (uid_t uid);
void rule_gen_lport (u_int16_t low, u_int16_t high);
void rule_gen_rport (u_int16_t low, u_int16_t high);
void rule_gen_laddr (u_int32_t addr, u_int32_t mask);
void rule_gen_raddr (u_int32_t addr, u_int32_t mask);
void rule_gen_or (void);
void rule_gen_and (void);
void rule_gen_not (void);
int rule_eval (uid_t r_uid, u_int32_t laddr, u_int16_t lport, 
		u_int32_t raddr, u_int16_t rport);
void rule_parse (const char *r);

/*
 * Internal helper functions.
 */
static void push (bool_t b);
static bool_t pop (void);
static void add_code (bytecode_t c);
static bytecode_t next_code (size_t *ip);

/*
 * Module-level variables.
 *
 * TODO: Put these inside the functions that use them.
 */

#define CODE_SIZE_STEP		512
#define STACK_SIZE_STEP		512

static bytecode_t *code = NULL;
static size_t code_size = 0;
static bool_t *stack = NULL;
static size_t stack_size = 0;
static size_t stack_ptr, code_length;

/*
 * Bytecode instructions.
 */
#define BC_NONE		0
#define BC_USER		1
#define BC_LPORT	2
#define BC_RPORT	3
#define BC_LADDR	4
#define BC_RADDR	5
#define BC_OR		64
#define BC_AND		65
#define BC_NOT		66

/*
 * rule_gen_user()
 *
 * Generate a user comparison.
 */
void rule_gen_user (uid_t uid)
{
	add_code (BC_USER);
	add_code ((bytecode_t) uid);
}

/*
 * rule_gen_lport()
 *
 * Generate a local port comparison.
 */
void rule_gen_lport (u_int16_t low, u_int16_t high)
{
	add_code (BC_LPORT);
	add_code ((bytecode_t) low);
	add_code ((bytecode_t) high);
}

/*
 * rule_gen_rport()
 *
 * Generate a remote port comparison.
 */
void rule_gen_rport (u_int16_t low, u_int16_t high)
{
	add_code (BC_RPORT);
	add_code ((bytecode_t) low);
	add_code ((bytecode_t) high);
}

void rule_gen_laddr (u_int32_t addr, u_int32_t mask)
{
	add_code (BC_LADDR);
	add_code ((bytecode_t) addr);
	add_code ((bytecode_t) mask);
}

void rule_gen_raddr (u_int32_t addr, u_int32_t mask)
{
	add_code (BC_RADDR);
	add_code ((bytecode_t) addr);
	add_code ((bytecode_t) mask);
}

/*
 * rule_gen_or()
 *
 * Generate a logical or.
 */
void rule_gen_or (void)
{
	add_code (BC_OR);
}

/*
 * rule_gen_and()
 *
 * Generate a logical and.
 */
void rule_gen_and (void)
{
	add_code (BC_AND);
}

/*
 * rule_gen_not()
 *
 * Generate a logical not.
 */
void rule_gen_not (void)
{
	add_code (BC_NOT);
}

/*
 * rule_eval()
 *
 * Execute the rule on the provided connection information.
 *
 * Returns nonzero if the connection matches the rule, zero otherwise.
 */
int rule_eval (uid_t muid, u_int32_t mladdr, u_int16_t mlport, 
		u_int32_t mraddr, u_int16_t mrport)
{
	size_t ip = 0;
	unsigned int c;

	stack_ptr = 0;

	for (ip = 0; ip < code_length; ) {
		c = next_code (&ip);
		switch (c) {

			/*
			 * Comparisons.
			 */

			case BC_USER:
				/*
				 * User comparison.
				 */
				{
				uid_t uid;

				uid = (uid_t) next_code (&ip);

				push ((uid == muid) ? 1 : 0);
				}
				break;
			case BC_LPORT:
				/*
				 * Local port comparison.
				 */
				{
				u_int16_t hiport, loport;

				loport = (u_int16_t) next_code (&ip);
				hiport = (u_int16_t) next_code (&ip);

				push (((mlport >= loport) && (mlport <= hiport))
						? 1 : 0);
				}
				break;
			case BC_RPORT:
				/*
				 * Remote port comparison.
				 */
				{
				u_int16_t hiport, loport;

				loport = (u_int16_t) next_code (&ip);
				hiport = (u_int16_t) next_code (&ip);

				push (((mrport >= loport) && (mrport <= hiport))
						? 1 : 0);
				}
				break;
			case BC_LADDR:
				/*
				 * Local address comparison.
				 */
				{
				u_int32_t addr, mask;

				addr = (u_int32_t) next_code (&ip);
				mask = (u_int32_t) next_code (&ip);

				push (((mladdr & mask) == addr) ? 1 : 0);
				}
				break;	
			case BC_RADDR:
				/*
				 * Remote address comparison.
				 */
				{
				u_int32_t addr, mask;

				addr = (u_int32_t) next_code (&ip);
				mask = (u_int32_t) next_code (&ip);

				push (((mraddr & mask) == addr) ? 1 : 0);
				}
				break;	
				
			/*
			 * Logical operations.
			 */

			case BC_OR:
				/*
				 * Logical or.
				 */
				{
				bool_t a, b;
				
				a = pop (); b = pop ();
				push (((a == 1) || (b == 1)) ? 1 : 0);
				}
				break;
			case BC_AND:
				/*
				 * Logical and.
				 */
				{
				bool_t a, b;

				a = pop (); b = pop ();
				push (((a == 1) && (b == 1)) ? 1 : 0);
				}
				break;
			case BC_NOT:
				/*
				 * Logical not.
				 */
				{
				bool_t a;

				a = pop ();
				push ((a == 0) ? 1 : 0);
				}
				break;

			default:
				/*
				 * Unhandled bytecode instruction, choke.
				 */
				assert (0);
		}
	}

	assert (stack_ptr == 1);
	return pop();
}

/*
 * rule_parse()
 *
 * Parse and compile the specified rule.
 */
void rule_parse (const char *r)
{
	extern int ruleparse (void);
	extern void rule_lexer_scan_string (const char *r);
	
	assert (r != NULL);
	rule_lexer_scan_string (r);
	if (ruleparse () != 0)
		abort ();
}

/*
 * add_code()
 *
 * Add a bytecode or piece of data to the code array, dynamically resizing the
 * array as necessary.
 */
static void add_code (bytecode_t c)
{
	if ((code_length + 1) > code_size) {
		bytecode_t *p;
		
		code_size += CODE_SIZE_STEP;
		if ((p = realloc (code, sizeof (*p) * code_size)) == NULL) {
			fprintf (stderr, "tcpspy: memory exhausted\n");
			exit (EXIT_FAILURE);
		}
		code = p;
		assert ((code_length + 1) <= code_size);
	}
	code[code_length++] = c;
}

/*
 * next_code()
 *
 * Return the next bytecode or piece of data.
 */
static bytecode_t next_code (size_t *ip)
{
	assert (ip != NULL);
	assert (*ip < code_length);
	return code[(*ip)++];
}

/*
 * push()
 *
 * Push a boolean value to the stack, dynamically resizing as necessary.
 */
static void push (bool_t b)
{
	if ((stack_ptr + 1) > stack_size) {
		bool_t *p;

		stack_size += STACK_SIZE_STEP;
		if ((p = realloc (stack, sizeof (*p) * stack_size)) == NULL) {
			fprintf (stderr, "tcpspy: memory exhausted\n");
			exit (EXIT_FAILURE);
		}
		stack = p;

		assert ((stack_ptr + 1) <= stack_size);
	}
	
	stack[stack_ptr++] = b;
}

/*
 * pop()
 *
 * Pop and return a boolean value from the stack.
 */
static bool_t pop (void)
{
	if (stack_ptr == 0) {
		fprintf (stderr, "tcpspy: rule data stack underflow\n");
		exit (EXIT_FAILURE);
	}
	
	return stack[--stack_ptr];
}
